# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django import forms
from django.utils import timezone
from django.apps import apps

from . import conf
from . import utils


class ConfigForm(forms.Form):
    """
    Base class for every registered config form.\
    It behaves like a regular form.

    Inherits from :py:class:`django.forms.Form`.\
    The :py:attr:`initial` attr will be updated with the\
    config values if any.

    All form fields implementing this, should have a unique\
    name to avoid clashing with other registered forms,\
    prefixing them with the app name is a good practice.

    :param \*args: Positional parameters passed to parent class
    :param \*\*kwargs: Keyword parameters passed to parent class
    """
    def __init__(self, *args, **kwargs):
        super(ConfigForm, self).__init__(*args, **kwargs)

        self.initial.update({
            field_name: getattr(conf.config, field_name)
            for field_name in self.fields
            if hasattr(conf.config, field_name)
        })

    def save(self):
        """
        Save the config with the cleaned data,\
        update the last modified date so\
        the config is reloaded on other process/nodes.\
        Reload the config so it can be called right away.
        """
        assert self.__class__ in conf.config._registry,\
            '%(class_name)s is not registered' % {
                'class_name': self.__class__.__name__
            }

        data = self.cleaned_data
        data['_updated_at'] = timezone.now()
        ConfigModel = apps.get_model('djconfig.Config')

        for field_name, value in data.items():
            value = utils.serialize(value)

            # TODO: use update_or_create
            count = ConfigModel.objects\
                .filter(key=field_name)\
                .update(value=value)

            if not count:
                ConfigModel.objects.create(key=field_name, value=value)

        conf.config._reload()



# -*- coding: utf-8 -*-

from __future__ import unicode_literals

from django.apps import apps


class Config(object):
    """
    Contain registry of config forms and\
    cache of key-value matching the forms field-value.

    All methods must be private to avoid clashing\
    with the dynamic attributes.

    This should be usually accessed through :py:data:`config`
    """
    def __init__(self):
        self._registry = set()
        self._cache = {}

    def __getattr__(self, key):
        """
        Map cache data to config attributes

        :return: The cache value for the accessed key/attribute
        """
        try:
            return self._cache[key]
        except KeyError:
            raise AttributeError('Attribute "%s" not found in config.' % key)

    def _register(self, form_class, check_middleware=True):
        """
        * Public

        Register a config form into the registry

        :param object form_class:\
        The form class to register, instance of :py:class:`djconfig.forms.ConfigForm`
        :param bool check_middleware: Check if\
        :py:class:`djconfig.middleware.DjConfigMiddleware`\
        is registered into ``settings.MIDDLEWARE_CLASSES``. Default True
        """
        from . import forms  # avoids circular dependency

        assert issubclass(form_class, forms.ConfigForm), \
            "The form does not inherit from ConfigForm"

        self._registry.add(form_class)

        if check_middleware:
            self._check_backend()

    @staticmethod
    def _check_backend():
        """
        * Private

        Check if :py:class:`djconfig.middleware.DjConfigMiddleware`\
        is registered into ``settings.MIDDLEWARE_CLASSES``
        """
        from django.conf import settings

        middlewares = set(settings.MIDDLEWARE_CLASSES)

        # Deprecated alias
        if "djconfig.middleware.DjConfigLocMemMiddleware" in middlewares:
            return

        if "djconfig.middleware.DjConfigMiddleware" in middlewares:
            return

        raise ValueError(
            "djconfig.middleware.DjConfigMiddleware is required "
            "but it was not found in MIDDLEWARE_CLASSES"
        )

    def _reload(self):
        """
        * Private

        Gets every registered form's field value.\
        If a field name is found in the db, it will load it from there.\
        Otherwise, the initial value from the field form is used
        """
        ConfigModel = apps.get_model('djconfig.Config')
        cache = {}
        data = dict(
            ConfigModel.objects
                .all()
                .values_list('key', 'value')
        )

        for form_class in self._registry:
            form = form_class(data=data)
            form.is_valid()

            initial = {
                field_name: field.initial
                for field_name, field in form.fields.items()
            }
            cache.update(initial)

            cleaned_data = {
                field_name: value
                for field_name, value in form.cleaned_data.items()
                if field_name in data
            }
            cache.update(cleaned_data)

        cache['_updated_at'] = data.get('_updated_at')
        self._cache = cache

    def _reload_maybe(self):
        """
        * Public

        Reload the config if the config\
        model has been updated. This is called\
        once on every request by the middleware.\
        Should not be called directly.
        """
        ConfigModel = apps.get_model('djconfig.Config')

        data = dict(
            ConfigModel.objects
                .filter(key='_updated_at')
                .values_list('key', 'value')
        )

        # Load for the first time
        if not hasattr(self, '_updated_at'):
            self._reload()

        if data.get('_updated_at') != self._updated_at:
            self._reload()

    # Unit test helpers
    def _reset(self):
        self._registry = set()
        self._cache = {}

    def _set(self, key, value):
        self._cache[key] = value

    def _set_many(self, items):
        self._cache.update({
            key: value
            for key, value in items.items()
        })


config = Config()

# Public methods
register = config._register
reload_maybe = config._reload_maybe



# -*- coding: utf-8 -*-

from __future__ import unicode_literals
from functools import wraps

from django.db import models

from . import conf


__all__ = [
    'override_djconfig',
    'serialize'
]


def override_djconfig(**new_cache_values):
    """
    Temporarily override config values.

    This is similar to :py:func:`django.test.override_settings`,\
    use it in testing.

    :param \*\*new_cache_values: Keyword arguments,\
    the key should match one in the config,\
    a new one is created otherwise,\
    the value is overridden within\
    the decorated function
    """
    def decorator(func):
        @wraps(func)
        def func_wrapper(*args, **kw):
            old_cache_values = {
                key: getattr(conf.config, key)
                for key in new_cache_values
            }

            conf.config._set_many(new_cache_values)

            try:
                # todo: make a note about this in the docs: don't populate the config within migrations

                # This works coz the config table is empty,
                # so even if the middleware gets called,
                # it won't update the config (_updated_at
                # will be None), this is assuming the table
                # is not populated by the user (ie: within
                # a migration), in which case it will load
                # all the default values
                return func(*args, **kw)
            finally:
                conf.config._set_many(old_cache_values)

        return func_wrapper

    return decorator


def serialize(value):
    """
    Form values serialization

    :param object value: A value to be serialized\
    for saving it into the database and later\
    loading it into the form as initial value
    """
    # todo: add DateField

    if isinstance(value, models.Model):
        return value.pk

    return value
